from typing import List, Tuple, Any
from sklearn.metrics import confusion_matrix as cm
import plotly.express as px
import matplotlib.pyplot as plt
import numpy as np


def relative_values(array: Any) -> np.ndarray:
    """ Function to calculate values relative to sum of array. """

    return array / sum(array)


def plot_training_process(loss_list: List, accuracy_list: List, figsize: Tuple = (12, 6)):
    """ Plot validation accuracy and validation loss over epochs.

    :param loss_list: List of validation loss by epoch.
    :param accuracy_list: List of validation loss by epoch.
    :param figsize: Tuple of figure size.
    """

    fig, (ax1, ax2) = plt.subplots(2, figsize=figsize, sharex=True)

    ax1.plot(accuracy_list)
    ax1.set_ylabel("validation accuracy")
    ax2.plot(loss_list)
    ax2.set_ylabel("validation loss")
    ax2.set_xlabel("epochs");


def create_heatmap(data: Any, classes: Any, colorscale: Any, title: str, height: int, width: int,
                   x_label: str = 'Predicted',
                   y_label: str = 'Actual'):
    """ Create a heatmap plot of given data and class labels.

    :param data: Heatmap values.
    :param classes: Class labels.
    :param colorscale: Plotly continuous colorscale.
    :param title: Figure title.
    :param height: Figure height.
    :param width: Figure width.
    :return: Plotly figure.
    """

    fig = px.imshow(data,
                    labels=dict(x=x_label, y=y_label),
                    x=classes,
                    y=classes,
                    color_continuous_scale=colorscale)

    fig.update_traces(text=data, texttemplate="%{text:.2f}")
    fig.update_layout(title_text=title)
    fig.layout.height = height
    fig.layout.width = width

    return fig


def confusion_heatmap(y_test: Any, y_pred: Any, classes: Any, colorscale='Blues', height: int = 900, width: int = 900,
                      title: str = 'Confusion heatmap'):
    """ Create heatmap plot based on confusion matrix of predicted and true values for arbitrary number of classes. Uses create_heatmap() function for plot generation.

    :param y_test: Actual class values of test data.
    :param y_pred: Predicted class values of test data.
    :param classes: Class labels.
    :param colorscale: Plotly colorscale.
    :param height: Figure height.
    :param width: Figure width.
    :param title: Figure title.
    :return: Plotly figure.
    """
    confusion_matrix = cm(y_test, y_pred)
    fig = create_heatmap(confusion_matrix, classes, colorscale, title, height, width)

    return fig


def precision_heatmap(y_test: Any, y_pred: Any, classes: Any, colorscale='Blues',
                      height: int = 900, width: int = 900,
                      title: str = 'Precision heatmap (column wise relative values)'):
    """ Create heatmap plot based on confusion matrix of predicted and true values with relative values along axis of each
    predicted class (column wise). This represents a precision value for each class in the diagonal of the heatmap. Uses create_heatmap() function for plot generation.

    :param y_test: Actual class values of test data.
    :param y_pred: Predicted class values of test data.
    :param classes: Class labels.
    :param colorscale: Plotly colorscale.
    :param height: Figure height.
    :param width: Figure width.
    :param title: Figure title.
    :return: Plotly figure.
    """

    confusion_matrix = cm(y_test, y_pred)
    confusion_matrix = np.apply_along_axis(relative_values, 0, confusion_matrix)
    fig = create_heatmap(confusion_matrix, classes, colorscale, title, height, width)

    return fig


def recall_heatmap(y_test: Any, y_pred: Any, classes: Any, colorscale='Blues',
                   height: int = 900, width: int = 900, title: str = 'Recall heatmap (row wise relative values)'):
    """ Create heatmap plot based on confusion matrix of predicted and true values with relative values along axis of each
    predicted class (row wise). This represents a recall value for each class in the diagonal of the heatmap. Uses create_heatmap() function for plot generation.

    :param y_test: Actual class values of test data.
    :param y_pred: Predicted class values of test data.
    :param classes: Class labels.
    :param colorscale: Plotly colorscale.
    :param height: Figure height.
    :param width: Figure width.
    :param title: Figure title.
    :return: Plotly figure.
    """
    confusion_matrix = cm(y_test, y_pred)
    confusion_matrix = np.apply_along_axis(relative_values, 1, confusion_matrix)
    fig = create_heatmap(confusion_matrix, classes, colorscale, title, height, width)

    return fig



import math
import random
import numpy as np
from numpy.random import default_rng
import torch
from sklearn.preprocessing import RobustScaler
from torch.autograd import Variable
from typing import Tuple, Any, List


def create_dataset(X: Any, y: Any, time_steps: int = 1, step_size: int = 1) -> Tuple[np.ndarray, np.ndarray]:
    """Function to split data with a categorical y variable into windows with length of given time_steps and an interval
    equal to value of step between the windows. For overlapping windows use a value for step smaller than time_steps.

    :param X: Dataframe with data of predictors.
    :param y: Dataframe column of target variable.
    :param time_steps: Length of windows in number of rows.
    :param step_size: Steps between windows.
    :return: Tuple of numpy array for x-variable data and numpy array for y-variable data."""

    X_windows, y_values = [], []
    for i in range(0, len(X) - time_steps, step_size):
        values = X.iloc[i:(i + time_steps)].values
        labels = y.iloc[i: i + time_steps]
        X_windows.append(values)
        y_values.append(labels.mode()[0])
    return np.array(X_windows), np.array(y_values).reshape(-1, 1)


def timeseries_train_test_split(data, x_variables, y_variable, splits: int, test_size: float, time_steps: int,
                                step_size: int, scale: bool = False) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Function to split a timeseries into windowed training and testing data without having data leakage even if
    using overlapping windows.

    :param data: Full dataframe with x and y variables.
    :param x_variables: List of column names of predictor variables.
    :param y_variable: Column name of target variable.
    :param splits: Number of random partitions in which data will be separated before windowing will be applied.
    :param test_size: Proportion of data to use for testing.
    :param time_steps: Length of windows in number of rows.
    :param step_size: Steps between windows.
    :param scale: Apply robust scaling.
    :return: Windowed arrays for training and testing data of predictors and target variable."""

    split_list = np.array_split(data, splits)

    # generate random numbers for selection of data splits
    rng = default_rng()
    random_numbers = rng.choice(splits, size=math.floor(splits * test_size), replace=False)

    df_full = data.copy()
    df_training = data.copy()

    # remove random partitions of data from training data
    for num in random_numbers:
        df_training = df_training.drop(split_list[num].index)

    if scale:
        # scale data based on training data
        scaler = RobustScaler()
        scaler = scaler.fit(df_training[x_variables].values)
        df_training.loc[:, x_variables] = scaler.transform(df_training[x_variables].to_numpy())

    # sample test splits
    X_test, y_test = [], []
    for num in random_numbers:
        split_idx = split_list[num].index

        # get data of partition from full dataframe
        extract = df_full.loc[split_idx, :]

        if scale:
            # apply scaling
            extract.loc[:, x_variables] = scaler.transform(extract[x_variables].to_numpy())

        # apply windowing function on test partitions
        X_slice, y_slice = create_dataset(extract[x_variables], extract[y_variable], time_steps, step_size)

        # append to test data
        X_test.append(X_slice)
        y_test.append(y_slice)

    X_test, y_test = np.vstack(X_test), np.vstack(y_test)

    # apply windowing function on training data
    X_train, y_train = create_dataset(df_training[x_variables], df_training[y_variable], time_steps, step_size)

    return X_train, y_train, X_test, y_test


def check_proportion_value(value: float):
    """ Function to check input of variable proportion_to_remove in function remove_maneuvers."""

    if (value > 1.0) or (value < 0):
        raise ValueError(f'Proportion to remove was set to {value} - has to be between 1.0 and 0.')


def remove_maneuvers(X_train: Any, y_train: Any, X_test: Any, y_test: Any, maneuvers: List[str],
                     proportion_to_remove: float = 1.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Function to remove maneuvers from given data partitions in a given proportion to balance
    maneuver classes.

    :param X_train: X data of training partition.
    :param y_train: Target variable data of training partition.
    :param X_test: X data of testing partition.
    :param y_test: Target variable data of testing partition.
    :param maneuvers: List of maneuver names or single maneuver name.
    :param proportion_to_remove: Proportion of data to remove for given maneuvers.
    :return: Numpy arrays of X and y data for training and testing partitions.
    """

    check_proportion_value(proportion_to_remove)

    if isinstance(maneuvers, str):
        maneuvers = [maneuvers]

    for maneuver in maneuvers:
        maneuver_train = np.where(y_train == maneuver)[0]
        maneuver_test = np.where(y_test == maneuver)[0]

        # randomly select cases to remove from training data
        remove_maneuver_train = random.sample(list(maneuver_train),
                                              k=round(proportion_to_remove * len(maneuver_train)))

        # randomly select cases to remove from training data
        remove_maneuver_test = random.sample(list(maneuver_test),
                                             k=round(proportion_to_remove * len(maneuver_test)))

        # delete selected cases from data
        X_train = np.delete(X_train, remove_maneuver_train, axis=0)
        y_train = np.delete(y_train, remove_maneuver_train)
        X_test = np.delete(X_test, remove_maneuver_test, axis=0)
        y_test = np.delete(y_test, remove_maneuver_test)

    return X_train, y_train, X_test, y_test


class LabelEncoding:
    """ Class to encode labels from target variable data and transform given data with sklearn LabelEncoder.
    Stores LabelEncoder for usage of inverse_transform(). """

    def __init__(self, y_train: Any, y_test: Any):
        """ Initialize LabelEncoding object based on sklearn LabelEncoder fitted on y training data.
        Attribute self.label_encoder can be used with function inverse_transform to transform y data back to encoded labels.

        :param y_train: Target variable data of training partition.
        :param y_test: Target variable data of testing partition.
        """
        from sklearn.preprocessing import LabelEncoder

        self.y_train = y_train
        self.y_test = y_test
        self.label_encoder = LabelEncoder()
        self.label_encoder.fit_transform(self.y_train)
        self.encoded_labels = self.label_encoder.classes_

    def transform(self):
        """ Function to transform y data by applying label encoding.
        :return: Transformed numpy arrays of training and testing y data.
        """
        self.y_train = self.label_encoder.transform(self.y_train)
        self.y_test = self.label_encoder.transform(self.y_test)

        return self.y_train, self.y_test


def transform_to_variables(X_train: Any, y_train: Any, X_test: Any, y_test: Any):
    """ Function to transform arrays of training and testing data to torch Variables.

    :param X_train: Numpy array of training predictor data.
    :param y_train: Numpy array of training target variable data.
    :param X_test: Numpy array of testing predictor data.
    :param y_test: Numpy array of training target variable data.
    :return: Data of type torch.Tensor for training and testing partitions.
    """

    X_train = Variable(torch.from_numpy(X_train)).float()
    y_train = Variable(torch.from_numpy(y_train)).long()
    X_test = Variable(torch.from_numpy(X_test)).float()
    y_test = Variable(torch.from_numpy(y_test)).long()
    return X_train, y_train, X_test, y_test



import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.utils.data import DataLoader


class ManeuverModel(nn.Module):
    """ Class with LSTM based model architecture for maneuver recognition."""

    def __init__(self, n_features: int, n_classes: int, n_hidden: int = 24,
                 n_layers: int = 4, lstm_dropout: float = 0.7,
                 n_features_linear1: int = 64, n_features_linear2: int = 32,
                 linear_dropout: float = 0.3):
        """ Initialization of model architecture.

        :param n_features: The number of expected features in the input x.
        :param n_classes: Number of classes for classification layer.
        :param n_hidden: Number of features in the hidden state of the LSTM.
        :param n_layers: Number of stacked LSTM layers.
        :param lstm_dropout: Value of applied dropout in LSTM layers.
        :param n_features_linear1: Number of features in first linear layer.
        :param n_features_linear2: Number of features in second linear layer.
        :param linear_dropout: Value of applied dropout between first and second linear layer.
        """

        super().__init__()

        self.lstm = nn.LSTM(
            input_size=n_features,
            hidden_size=n_hidden,
            num_layers=n_layers,
            batch_first=True,
            dropout=lstm_dropout
        )

        self.full_layer1 = nn.Linear(n_hidden, n_features_linear1)
        self.dropout = nn.Dropout(linear_dropout)
        self.full_layer2 = nn.Linear(n_features_linear1, n_features_linear2)
        self.classifier = nn.Linear(n_features_linear2, n_classes)

    def forward(self, x):
        """ Forward propagation. """
        self.lstm.flatten_parameters()

        _, (hidden, _) = self.lstm(x)

        out = hidden[-1]
        out = self.dropout(F.relu(self.full_layer1(out)))
        out = F.relu(self.full_layer2(out))
        out = self.classifier(out)
        return out

    def predict(self, X):
        """ Function to use model for prediction of given cases.

        :param X: X data to use for prediction.
        :return: List of predictions for given input data.
        """

        return [pred.argmax() for pred in self(X)]


def train(dataloader, model, loss_fn, optimizer, device):
    """ Function to apply training process on model with given data of dataloader object.
    In order to fit the model with direct data use fit_model.

    :param dataloader: Dataloader object for training data.
    :param model: Model object.
    :param loss_fn: Loss function.
    :param optimizer: Optimizer object for optimization algorithm.
    :param device: Device to use.
    """

    model.train()
    for batch, (X, y) in enumerate(dataloader):
        X, y = X.to(device), y.to(device)

        # Compute prediction error
        pred = model(X)
        loss = loss_fn(pred, y)

        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if batch % 100 == 0:
            loss, current = loss.item(), batch * len(X)


def test(dataloader, model, loss_fn, device):
    """ Function to evaluate given model with data of dataloader. In order to use the model for predictions use
    the predict function of the model object instead.

    :param dataloader: Dataloader object for test data.
    :param model: Model object.
    :param loss_fn: Loss function.
    :param device: Device to use.
    :return: Tuple with proportion of correctly predicted cases and average test loss.
    """

    size = len(dataloader.dataset)
    num_batches = len(dataloader)
    model.eval()
    test_loss, correct = 0, 0
    with torch.no_grad():
        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            pred = model(X)
            test_loss += loss_fn(pred, y).item()
            correct += (pred.argmax(1) == y).type(torch.float).sum().item()
    test_loss /= num_batches
    correct /= size
    print(f"Test Error: Accuracy: {(100 * correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")

    return [correct, test_loss]


def fit_model(model, X_train, y_train, X_test, y_test, epochs, batch_size,
                         loss_function, optimizer, device):
    """ Function to fit a model. Applies model training with given X and y training data and uses given X and y test
    data for training validation. Returns list of validation loss and validation accuracy per epoch.

    :param model: Model object
    :param X_train: X data of training partition.
    :param y_train: Target variable data of training partition.
    :param X_test: X data of testing partition.
    :param y_test: Target variable data of testing partition.
    :param epochs: Number of epochs in training process.
    :param batch_size: Batch size to use for training and testing.
    :param loss_function: Loss function.
    :param optimizer: Optimizer object for optimization algorithm.
    :param device: Device to use.
    :return: List of validation loss and validation accuracy values for every epoch.
    """

    train_dataloader = DataLoader(list(zip(X_train, y_train)), batch_size=batch_size)
    test_dataloader = DataLoader(list(zip(X_test, y_test)), batch_size=batch_size)

    loss_list = np.zeros((epochs,))
    accuracy_list = np.zeros((epochs,))

    for epoch in range(epochs):
        print(f"Epoch {epoch + 1} / {epochs}\n-------------------------------")
        train(train_dataloader, model, loss_function, optimizer, device)
        accuracy, test_loss = test(test_dataloader, model, loss_function, device)
        accuracy_list[epoch] = accuracy
        loss_list[epoch] = test_loss
    print("Done!")

    return loss_list, accuracy_list
